/*******************************************************************************
 * Copyright (C) 2010-2018 Dominik Egert <dominik.egert@freie-netze.de>
 * 
 * 
 * This file is Public Domain licensed and can be copied and/or distributed 
 * at free will.
 **********************************************************************************
 *  
 *  File information
 *  ================
 *  
 *  Use: Helper utilities to support easy data handling
 *  
 *   Exception: doubles. Due to their nature, 
 *   and the absolute irresponsible fact of using doubles for 
 *   representing monetary values, they need "special treatment".
 * 
 *   This has been realised (as far as possible, mathematically).
 *   Allthough the class itself is no double but an object, most 
 *   operations can be done and find their corresponding overwrite 
 *   in the derived classes.
 *  
 * 
 *  This is all about data handling!!
*/


// Representant for the most recent fully compatible basic double class 
#define dbl2 ea_dbl_ext
#define dblm ea_dbl_magnet



/***************************************************************************************/
/* Double value handler class                                                          */
/***************************************************************************************/

//
// Class ea_dbl + ea_dbl_ext
//  Used to handle odds of double values.
//  provides access to ticks, pips and points defined by market digits
//  handles all types of numeric functionality in a consistent way 
//  always providing full double value access
//
//   dbl_ext as a provider for more specific overwrites of operators
//   dbl_magnet as a provider for different comparison operator 
//
//  LIMITATION:
//    The maximum consistent amount of digits is set to 18.
//      So the maximum hex value is set to: 0x0DE0 B6B3 A763 FFFF  (dec: 999.999.999.999.999.999)
//      And the minimum hex value is set to: 0xF21F 494C 589C 0001 (dec: -999.999.999.999.999.999)
// 
//		This limitation is not being checked. So the internal long-variable 
//      will experience an overflow.
//
// 
#define DBL2_MAX 0x0DE0B6B3A763FFFF
#define DBL2_MIN 0xF21F494C589C0001


// Basic double representation class
class ea_dbl
{
    // Internal storage
    private:
        double  m_dbl;
        double  m_rounding;
        bool    m_bPRound;
        bool    m_bEmpty;
        long    m_long;
        long    m_precision;
        long    m_saved;
        long    m_points;
        long    m_pips;
        long    m_ticks;


    // Object handlers
    public:
                void    ea_dbl()                                                 { m_dbl = 0; m_saved = Digits() + 3; SetPrecision((int)m_saved); SetRounding(); m_bEmpty = true; }
                void    ea_dbl(const double inVal)                               { m_dbl = inVal; m_saved = Digits() + 3; SetPrecision((int)m_saved); SetRounding(); } 
                void    ea_dbl(const ea_dbl& inVal)                              { m_dbl = inVal.Get(); m_saved = inVal.GetDigits(); SetPrecision((int)m_saved); SetRounding(inVal.GetRounding()); }
                void    ~ea_dbl()                                                { return; }
                                                                          
    
	
    // Public interface                                                   
    public:
	
		// Virtual interface
        virtual double  Get()                                              const { return(GetRawDouble()); }
        virtual void    SetRounding(const bool bPrecise = false)                 { double r = 0; m_bPRound = bPrecise; if(m_bPRound) { r = (4.9 / ((double)m_points*10)); } m_rounding = r; }
        virtual bool    GetRounding()                                      const { return(m_bPRound); }

		
		// Default interface
                void    SetPrecision(const int inVal = 8)                        {  m_precision = (long)inVal; 
                                                                                    m_points = (long)MathPow(10, m_precision); 
                                                                                    m_pips = (long)MathPow(10, (m_precision + 1) - Digits()); 
                                                                                    m_ticks = (long)MathPow(10, m_precision - Digits()); update(); }
				long    GetLong()                                          const { return(m_long); }
                double  GetDouble()                                        const { return(to_double(m_long)); }
				double	GetRawDouble()									   const { return(m_dbl); } 
                long    GetDigits()                                        const { return(m_precision); }
                void    CopyPrecision(const ea_dbl &inVal)                       { SetPrecision((int)inVal.GetDigits()); } 
                                                                          

		// Sync provider (digit-sync between objects)                         
                void    _sync()                                                  { SetPrecision((int)m_saved); }
                long    _sync(ea_dbl &inVal)                                     { return(_sync(inVal._sync(m_precision))); }
                long    _sync(const long inVal)                                  { m_saved = m_precision; if(inVal < m_precision) { SetPrecision((int)inVal); } return(m_precision); }
                void    _unsync(ea_dbl &inVal)                                   { _sync(); inVal._sync(); }
                                                                          
                                                                          
		// Set and get different types of digits/values                       
                void    ticks(const long inVal)                                  { m_dbl = to_double((inVal*m_ticks)); update(); }
                long    ticks()                                            const { return(m_long/m_ticks); }
                void    pips(const long inVal)                                   { m_dbl = to_double((inVal*m_pips)); update(); }
                long    pips()                                             const { return(m_long/m_pips); }
                void    points(const long inVal)                                 { m_dbl = to_double((inVal*m_points)); update(); }
                long    points()                                           const { return(m_long/m_points); }
                                                                          

                                                                          
    // Internal helper functions                                          
    private:                                                              

		// Keep internal data updated
                void    update()                                                 {  m_bEmpty = false; m_long = to_long(m_dbl); return; }
                                                                          

    protected:                                                            
		// Helpers, securing correct digit count handling
                long    to_ticks(const long inVal)                         const { return(inVal/m_ticks); }
                long    to_pips(const long inVal)                          const { return(inVal/m_pips); }
                long    to_points(const long inVal)                        const { return(inVal/m_points); }
                long    from_ticks(const long inVal)                       const { return(inVal*m_ticks); }
                long    from_pips(const long inVal)                        const { return(inVal*m_pips); }
                long    from_points(const long inVal)                      const { return(inVal*m_points); }
                long    to_long(const double inVal)                        const {  string str = DoubleToString(MathRound((double)m_points * (inVal + m_rounding)), 0);
                                                                                   return( StringToInteger(str) ); }
                double  to_double(const long inVal)                        const {  string str = IntegerToString(inVal, (int)m_precision, 0x30);
                                                                                    bool bNeg = false; 
                                                                                    if(StringReplace(str, "-", "") > 0) { bNeg = true; }
                                                                                    string missing_zeroes = ""; 
                                                                                    for(int pos = StringLen(str) - (int)m_precision; pos <= 0; pos++)
                                                                                    { StringAdd(missing_zeroes, "0"); }
                                                                                    StringAdd(missing_zeroes, str);
                                                                                    StringConcatenate(str, bNeg?"-":"", StringSubstr(missing_zeroes, 0, StringLen(missing_zeroes) - (int)m_precision),
                                                                                                        ".", StringSubstr(missing_zeroes, StringLen(missing_zeroes) - (int)m_precision));
                                                                                   return(StringToDouble(str)); }
                                                                          
                                                                          
    // Operators
    public:                                                       
	
    // Generic Operators: Operators: !, ++, --                                                       
	
		// ea_dbl selfreferencing operators 
            virtual bool    operator !  ()                                 const { return((m_bEmpty) || (m_long == 0)); }
            virtual void    operator ++ ()                                       { m_dbl += 1.0/(MathPow(10, Digits())); update(); }
            virtual void    operator -- ()                                       { m_dbl -= 1.0/(MathPow(10, Digits())); update(); }
                                           
										   
    // Operators: ==, !=, <, >, <=, >=, +, -, *, /, %, =, +=, -=, *=, /=, %=
                                                                          
        // dbl class object reference                                     
            virtual bool    operator == (ea_dbl &inVal)                          { _sync(inVal); bool bEqual = (m_long == inVal.GetLong()); _unsync(inVal); return(bEqual); }
            virtual bool    operator != (ea_dbl &inVal)                          { return(!operator == (inVal)); }
            virtual bool    operator >  (ea_dbl &inVal)                          { _sync(inVal); bool bEqual = (m_long >  inVal.GetLong()); _unsync(inVal); return(bEqual); }
            virtual bool    operator <  (ea_dbl &inVal)                          { _sync(inVal); bool bEqual = (m_long <  inVal.GetLong()); _unsync(inVal); return(bEqual); }
            virtual bool    operator >= (ea_dbl &inVal)                          { _sync(inVal); bool bEqual = (m_long >= inVal.GetLong()); _unsync(inVal); return(bEqual); }
            virtual bool    operator <= (ea_dbl &inVal)                          { _sync(inVal); bool bEqual = (m_long <= inVal.GetLong()); _unsync(inVal); return(bEqual); }
                                                                              
                    ea_dbl  operator +  (ea_dbl &inVal)                          { _sync(inVal); long lRetval = (m_long + inVal.GetLong()); _unsync(inVal); return(to_double(lRetval)); }
                    ea_dbl  operator -  (ea_dbl &inVal)                          { _sync(inVal); long lRetval = (m_long - inVal.GetLong()); _unsync(inVal); return(to_double(lRetval)); }
                    ea_dbl  operator *  (ea_dbl &inVal)                          { _sync(inVal); long lRetval = (m_long * inVal.GetLong()); _unsync(inVal); return(to_double(lRetval)); }
                    ea_dbl  operator /  (ea_dbl &inVal)                          { _sync(inVal); long lRetval = (m_long / inVal.GetLong()); _unsync(inVal); return(to_double(lRetval)); }
                    ea_dbl  operator %  (ea_dbl &inVal)                          { _sync(inVal); long lRetval = (m_long % inVal.GetLong()); _unsync(inVal); return(to_double(lRetval)); }
                                                                              
                    void    operator =  (ea_dbl &inVal)                          { _sync(inVal); m_long =  inVal.GetLong(); m_dbl = to_double(m_long); _unsync(inVal); return; }
                    void    operator += (ea_dbl &inVal)                          { _sync(inVal); m_long += inVal.GetLong(); m_dbl = to_double(m_long); _unsync(inVal); return; }
                    void    operator -= (ea_dbl &inVal)                          { _sync(inVal); m_long -= inVal.GetLong(); m_dbl = to_double(m_long); _unsync(inVal); return; }
                    void    operator *= (ea_dbl &inVal)                          { _sync(inVal); m_long *= inVal.GetLong(); m_dbl = to_double(m_long); _unsync(inVal); return; }
                    void    operator /= (ea_dbl &inVal)                          { _sync(inVal); m_long /= inVal.GetLong(); m_dbl = to_double(m_long); _unsync(inVal); return; }
                    void    operator %= (ea_dbl &inVal)                          { ticks((ticks()%inVal.ticks())); }


        // double as input
            virtual bool    operator == (const double inVal)               const { return(m_long == to_long(inVal)); }
            bool            operator != (const double inVal)               const { return(!operator == (inVal)); }
            virtual bool    operator >  (const double inVal)               const { return(m_long >  to_long(inVal)); }
            virtual bool    operator <  (const double inVal)               const { return(m_long <  to_long(inVal)); }
            virtual bool    operator >= (const double inVal)               const { return(m_long >= to_long(inVal)); }
            virtual bool    operator <= (const double inVal)               const { return(m_long <= to_long(inVal)); }
                                                                          
                    double  operator +  (const double inVal)               const { return(m_dbl+inVal); }
                    double  operator -  (const double inVal)               const { return(m_dbl-inVal); }
                    double  operator *  (const double inVal)               const { return(m_dbl*inVal); }
                    double  operator /  (const double inVal)               const { return(m_dbl/inVal); }
                    double  operator %  (const double inVal)               const { return((double)((m_long/m_ticks)%(to_long(inVal)/m_ticks))); }
                                                                            
                    void    operator =  (const double inVal)                     { m_dbl =  inVal; update(); }
                    void    operator += (const double inVal)                     { m_dbl += inVal; update(); }
                    void    operator -= (const double inVal)                     { m_dbl -= inVal; update(); }
                    void    operator *= (const double inVal)                     { m_dbl *= inVal; update(); }
                    void    operator /= (const double inVal)                     { m_dbl /= inVal; update(); }
                    void    operator %= (const double inVal)                     { m_dbl = (double)((m_long/m_ticks)%(to_long(inVal)/m_ticks)); update(); }
};



//
// Operators overwrite class
//   
class ea_dbl_ext : public ea_dbl
{
    // Object handlers
    public:
        void    ea_dbl_ext() : ea_dbl()                                          { return; }
        void    ~ea_dbl_ext()                                                    { return; }
        void    ea_dbl_ext(const double inVal) : ea_dbl(inVal)                   { return; } 
        void    ea_dbl_ext(const ea_dbl_ext& inVal) : ea_dbl(inVal)              { return; }
        void    ea_dbl_ext(const string inVal)                                   { SetPrecision((int)digits_from_string(inVal)); SetRounding(); operator = (StringToDouble(inVal)); }
                                                                          
    // Internal helper                                                    
    private:                                                              
        int     digits_from_string(const string inVal)                     const {  int i; 
                                                                                    for(i = StringFind(inVal, ".") + 1; i < StringLen(inVal); i++)  
                                                                                    { if(inVal[i] != 0x30) { i++; break; } } 
                                                                                   return(i - (StringFind(inVal, ".") + 1)); }; 
                                                                          
                                                                          
    // Additional interface                                               
    public:                                                               
        void    ticks(const int inVal)                                           { ticks((long)inVal); }
        void    pips(const int inVal)                                            { pips((long)inVal); }
        void    points(const int inVal)                                          { points((long)inVal); }
                                                                          
        void    ticks(const int &inVal)                                          { ticks((long)inVal); }
        void    pips(const int &inVal)                                           { pips((long)inVal); }
        void    points(const int &inVal)                                         { points((long)inVal); }
                                                                          
        void    ticks(const long &inVal)                                         { ticks((long)inVal); }
        void    pips(const long &inVal)                                          { pips((long)inVal); }
        void    points(const long &inVal)                                        { points((long)inVal); }

    
    // Supported operators: ==, !=, <, >, <=, >=, +, -, *, /, %, =, +=, -=, *=, /=, %=
	// Supported input types: uint, int, long, ulong, &uint, &int, &ulong, &long, &double


    // Smart learn assignment
            void        operator =  (const string inVal)                         { SetPrecision((int)digits_from_string(inVal)); SetRounding(); operator = (StringToDouble(inVal)); }
                                                                          
    // Copy-Value as input                                                
        // uint as input                                                  
            bool        operator == (const uint inVal)                     const { return(operator == ((double)inVal)); }
            bool        operator != (const uint inVal)                     const { return(operator != ((double)inVal)); }
            bool        operator >  (const uint inVal)                     const { return(operator >  ((double)inVal)); }
            bool        operator <  (const uint inVal)                     const { return(operator <  ((double)inVal)); }
            bool        operator >= (const uint inVal)                     const { return(operator >= ((double)inVal)); }
            bool        operator <= (const uint inVal)                     const { return(operator <= ((double)inVal)); }
                                                                          
            double      operator +  (const uint inVal)                     const { return(operator +  ((double)inVal)); }
            double      operator -  (const uint inVal)                     const { return(operator -  ((double)inVal)); }
            double      operator *  (const uint inVal)                     const { return(operator *  ((double)inVal)); }
            double      operator /  (const uint inVal)                     const { return(operator /  ((double)inVal)); }
            double      operator %  (const uint inVal)                     const { return(operator %  ((double)inVal)); }
                                                                          
            void        operator =  (const uint inVal)                           { operator =  ((double)inVal); }
            void        operator += (const uint inVal)                           { operator += ((double)inVal); }
            void        operator -= (const uint inVal)                           { operator -= ((double)inVal); }
            void        operator *= (const uint inVal)                           { operator *= ((double)inVal); }
            void        operator /= (const uint inVal)                           { operator /= ((double)inVal); }
            void        operator %= (const uint inVal)                           { operator %= ((double)inVal); }
                                                                          
                                                                          
        // int as input                                                   
            bool        operator == (const int inVal)                      const { return(operator == ((double)inVal)); }
            bool        operator != (const int inVal)                      const { return(operator != ((double)inVal)); }
            bool        operator >  (const int inVal)                      const { return(operator >  ((double)inVal)); }
            bool        operator <  (const int inVal)                      const { return(operator <  ((double)inVal)); }
            bool        operator >= (const int inVal)                      const { return(operator >= ((double)inVal)); }
            bool        operator <= (const int inVal)                      const { return(operator <= ((double)inVal)); }
                                                                          
            double      operator +  (const int inVal)                      const { return(operator +  ((double)inVal)); }
            double      operator -  (const int inVal)                      const { return(operator -  ((double)inVal)); }
            double      operator *  (const int inVal)                      const { return(operator *  ((double)inVal)); }
            double      operator /  (const int inVal)                      const { return(operator /  ((double)inVal)); }
            double      operator %  (const int inVal)                      const { return(operator %  ((double)inVal)); }
                                                                          
            void        operator =  (const int inVal)                            { operator =  ((double)inVal); }
            void        operator += (const int inVal)                            { operator += ((double)inVal); }
            void        operator -= (const int inVal)                            { operator -= ((double)inVal); }
            void        operator *= (const int inVal)                            { operator *= ((double)inVal); }
            void        operator /= (const int inVal)                            { operator /= ((double)inVal); }
            void        operator %= (const int inVal)                            { operator %= ((double)inVal); }
                                                                          
                                                                          
        // ulong as input                                                 
            bool        operator == (const ulong inVal)                    const { return(operator == ((double)inVal)); }
            bool        operator != (const ulong inVal)                    const { return(operator != ((double)inVal)); }
            bool        operator >  (const ulong inVal)                    const { return(operator >  ((double)inVal)); }
            bool        operator <  (const ulong inVal)                    const { return(operator <  ((double)inVal)); }
            bool        operator >= (const ulong inVal)                    const { return(operator >= ((double)inVal)); }
            bool        operator <= (const ulong inVal)                    const { return(operator <= ((double)inVal)); }
                                                                          
            double      operator +  (const ulong inVal)                    const { return(operator +  ((double)inVal)); }
            double      operator -  (const ulong inVal)                    const { return(operator -  ((double)inVal)); }
            double      operator *  (const ulong inVal)                    const { return(operator *  ((double)inVal)); }
            double      operator /  (const ulong inVal)                    const { return(operator /  ((double)inVal)); }
            double      operator %  (const ulong inVal)                    const { return(operator %  ((double)inVal)); }
                                                                          
            void        operator =  (const ulong inVal)                          { operator =  ((double)inVal); }
            void        operator += (const ulong inVal)                          { operator += ((double)inVal); }
            void        operator -= (const ulong inVal)                          { operator -= ((double)inVal); }
            void        operator *= (const ulong inVal)                          { operator *= ((double)inVal); }
            void        operator /= (const ulong inVal)                          { operator /= ((double)inVal); }
            void        operator %= (const ulong inVal)                          { operator %= ((double)inVal); }
                                                                          
                                                                          
        // long as input                                                  
            bool        operator == (const long inVal)                     const { return(operator == ((double)inVal)); }
            bool        operator != (const long inVal)                     const { return(operator != ((double)inVal)); }
            bool        operator >  (const long inVal)                     const { return(operator >  ((double)inVal)); }
            bool        operator <  (const long inVal)                     const { return(operator <  ((double)inVal)); }
            bool        operator >= (const long inVal)                     const { return(operator >= ((double)inVal)); }
            bool        operator <= (const long inVal)                     const { return(operator <= ((double)inVal)); }
                                                                          
            double      operator +  (const long inVal)                     const { return(operator +  ((double)inVal)); }
            double      operator -  (const long inVal)                     const { return(operator -  ((double)inVal)); }
            double      operator *  (const long inVal)                     const { return(operator *  ((double)inVal)); }
            double      operator /  (const long inVal)                     const { return(operator /  ((double)inVal)); }
            double      operator %  (const long inVal)                     const { return(operator %  ((double)inVal)); }
                                                                          
            void        operator =  (const long inVal)                           { operator =  ((double)inVal); }
            void        operator += (const long inVal)                           { operator += ((double)inVal); }
            void        operator -= (const long inVal)                           { operator -= ((double)inVal); }
            void        operator *= (const long inVal)                           { operator *= ((double)inVal); }
            void        operator /= (const long inVal)                           { operator /= ((double)inVal); }
            void        operator %= (const long inVal)                           { operator %= ((double)inVal); }
                                                                          
                                                                          
    // References as input                                                
        // int referenceas input                                          
            bool        operator == (const uint &inVal)                    const { return(operator == ((double)inVal)); }
            bool        operator != (const uint &inVal)                    const { return(operator != ((double)inVal)); }
            bool        operator >  (const uint &inVal)                    const { return(operator >  ((double)inVal)); }
            bool        operator <  (const uint &inVal)                    const { return(operator <  ((double)inVal)); }
            bool        operator >= (const uint &inVal)                    const { return(operator >= ((double)inVal)); }
            bool        operator <= (const uint &inVal)                    const { return(operator <= ((double)inVal)); }
                                                                          
            double      operator +  (const uint &inVal)                    const { return(operator +  ((double)inVal)); }
            double      operator -  (const uint &inVal)                    const { return(operator -  ((double)inVal)); }
            double      operator *  (const uint &inVal)                    const { return(operator *  ((double)inVal)); }
            double      operator /  (const uint &inVal)                    const { return(operator /  ((double)inVal)); }
            double      operator %  (const uint &inVal)                    const { return(operator %  ((double)inVal)); }
                                                                          
            void        operator =  (const uint &inVal)                          { operator =  ((double)inVal); }
            void        operator += (const uint &inVal)                          { operator += ((double)inVal); }
            void        operator -= (const uint &inVal)                          { operator -= ((double)inVal); }
            void        operator *= (const uint &inVal)                          { operator *= ((double)inVal); }
            void        operator /= (const uint &inVal)                          { operator /= ((double)inVal); }
            void        operator %= (const uint &inVal)                          { operator %= ((double)inVal); }
                                                                          
                                                                          
        // int referenceas input                                          
            bool        operator == (const int &inVal)                     const { return(operator == ((double)inVal)); }
            bool        operator != (const int &inVal)                     const { return(operator != ((double)inVal)); }
            bool        operator >  (const int &inVal)                     const { return(operator >  ((double)inVal)); }
            bool        operator <  (const int &inVal)                     const { return(operator <  ((double)inVal)); }
            bool        operator >= (const int &inVal)                     const { return(operator >= ((double)inVal)); }
            bool        operator <= (const int &inVal)                     const { return(operator <= ((double)inVal)); }
                                                                          
            double      operator +  (const int &inVal)                     const { return(operator +  ((double)inVal)); }
            double      operator -  (const int &inVal)                     const { return(operator -  ((double)inVal)); }
            double      operator *  (const int &inVal)                     const { return(operator *  ((double)inVal)); }
            double      operator /  (const int &inVal)                     const { return(operator /  ((double)inVal)); }
            double      operator %  (const int &inVal)                     const { return(operator %  ((double)inVal)); }
                                                                          
            void        operator =  (const int &inVal)                           { operator =  ((double)inVal); }
            void        operator += (const int &inVal)                           { operator += ((double)inVal); }
            void        operator -= (const int &inVal)                           { operator -= ((double)inVal); }
            void        operator *= (const int &inVal)                           { operator *= ((double)inVal); }
            void        operator /= (const int &inVal)                           { operator /= ((double)inVal); }
            void        operator %= (const int &inVal)                           { operator %= ((double)inVal); }
                                                                          
                                                                          
        // ulong as input                                                 
            bool        operator == (const ulong &inVal)                   const { return(operator == ((double)inVal)); }
            bool        operator != (const ulong &inVal)                   const { return(operator != ((double)inVal)); }
            bool        operator >  (const ulong &inVal)                   const { return(operator >  ((double)inVal)); }
            bool        operator <  (const ulong &inVal)                   const { return(operator <  ((double)inVal)); }
            bool        operator >= (const ulong &inVal)                   const { return(operator >= ((double)inVal)); }
            bool        operator <= (const ulong &inVal)                   const { return(operator <= ((double)inVal)); }
                                                                          
            double      operator +  (const ulong &inVal)                   const { return(operator +  ((double)inVal)); }
            double      operator -  (const ulong &inVal)                   const { return(operator -  ((double)inVal)); }
            double      operator *  (const ulong &inVal)                   const { return(operator *  ((double)inVal)); }
            double      operator /  (const ulong &inVal)                   const { return(operator /  ((double)inVal)); }
            double      operator %  (const ulong &inVal)                   const { return(operator %  ((double)inVal)); }
                                                                          
            void        operator =  (const ulong &inVal)                         { operator =  ((double)inVal); }
            void        operator += (const ulong &inVal)                         { operator += ((double)inVal); }
            void        operator -= (const ulong &inVal)                         { operator -= ((double)inVal); }
            void        operator *= (const ulong &inVal)                         { operator *= ((double)inVal); }
            void        operator /= (const ulong &inVal)                         { operator /= ((double)inVal); }
            void        operator %= (const ulong &inVal)                         { operator %= ((double)inVal); }
                                                                          
                                                                          
        // long as input                                                  
            bool        operator == (const long &inVal)                    const { return(operator == ((double)inVal)); }
            bool        operator != (const long &inVal)                    const { return(operator != ((double)inVal)); }
            bool        operator >  (const long &inVal)                    const { return(operator >  ((double)inVal)); }
            bool        operator <  (const long &inVal)                    const { return(operator <  ((double)inVal)); }
            bool        operator >= (const long &inVal)                    const { return(operator >= ((double)inVal)); }
            bool        operator <= (const long &inVal)                    const { return(operator <= ((double)inVal)); }
                                                                          
            double      operator +  (const long &inVal)                    const { return(operator +  ((double)inVal)); }
            double      operator -  (const long &inVal)                    const { return(operator -  ((double)inVal)); }
            double      operator *  (const long &inVal)                    const { return(operator *  ((double)inVal)); }
            double      operator /  (const long &inVal)                    const { return(operator /  ((double)inVal)); }
            double      operator %  (const long &inVal)                    const { return(operator %  ((double)inVal)); }
                                                                          
            void        operator =  (const long &inVal)                          { operator =  ((double)inVal); }
            void        operator += (const long &inVal)                          { operator += ((double)inVal); }
            void        operator -= (const long &inVal)                          { operator -= ((double)inVal); }
            void        operator *= (const long &inVal)                          { operator *= ((double)inVal); }
            void        operator /= (const long &inVal)                          { operator /= ((double)inVal); }
            void        operator %= (const long &inVal)                          { operator %= ((double)inVal); }
                                                                          
                                                                          
        // double referenceas input                                       
            bool        operator == (const double &inVal)                  const { return(operator == ((double)inVal)); }
            bool        operator != (const double &inVal)                  const { return(operator != ((double)inVal)); }
            bool        operator >  (const double &inVal)                  const { return(operator >  ((double)inVal)); }
            bool        operator <  (const double &inVal)                  const { return(operator <  ((double)inVal)); }
            bool        operator >= (const double &inVal)                  const { return(operator >= ((double)inVal)); }
            bool        operator <= (const double &inVal)                  const { return(operator <= ((double)inVal)); }
                                                                          
            double      operator +  (const double &inVal)                  const { return(operator +  ((double)inVal)); }
            double      operator -  (const double &inVal)                  const { return(operator -  ((double)inVal)); }
            double      operator *  (const double &inVal)                  const { return(operator *  ((double)inVal)); }
            double      operator /  (const double &inVal)                  const { return(operator /  ((double)inVal)); }
            double      operator %  (const double &inVal)                  const { return(operator %  ((double)inVal)); }
                                                                          
            void        operator =  (const double &inVal)                        { operator =  ((double)inVal); }
            void        operator += (const double &inVal)                        { operator += ((double)inVal); }
            void        operator -= (const double &inVal)                        { operator -= ((double)inVal); }
            void        operator *= (const double &inVal)                        { operator *= ((double)inVal); }
            void        operator /= (const double &inVal)                        { operator /= ((double)inVal); }
            void        operator %= (const double &inVal)                        { operator %= ((double)inVal); }
};



//
// Comparison operators overwrite class enabling "magnet-mode"
//   
class ea_dbl_magnet : public ea_dbl_ext
{
    // Internal storage
    private:
        long    m_magnet;


    // Object handlers
    public:
        void    ea_dbl_magnet() : ea_dbl_ext()                                   { m_magnet = 0; SetRounding(true); }
        void    ea_dbl_magnet(const double inVal) : ea_dbl_ext(inVal)            { m_magnet = 0; SetRounding(true); } 
        void    ea_dbl_magnet(const ea_dbl_magnet& inVal) : ea_dbl_ext(inVal)    { m_magnet = inVal.GetMagnet_Precise(); }
        void    ~ea_dbl_magnet()                                                 { return; }


    // Public interface extenstion
    public:
        void    SetMagnet_Precise(const long inVal)                              { m_magnet = inVal; }
        void    SetMagnet(const double inVal)                                    { m_magnet = to_long(inVal); }
        void    SetMagnet_Ticks(const int inVal)                                 { m_magnet = from_ticks((long)inVal); }
        void    SetMagnet_Pips(const int inVal)                                  { m_magnet = from_pips((long)inVal); }
        void    SetMagnet_Points(const int inVal)                                { m_magnet = from_points((long)inVal); }
                                                                          
        long    GetMagnet_Precise()                                        const { return(m_magnet); }
        double  GetMagnet()                                                const { return(to_double(m_magnet)); }
        int     GetMagnet_Ticks()                                          const { return((int)to_ticks(m_magnet)); }
        int     GetMagnet_Pips()                                           const { return((int)to_pips(m_magnet)); }
        int     GetMagnet_Points()                                         const { return((int)to_points(m_magnet)); }
                                                                          
                                                                          
    // Internal helper functions                                          
    private:                                                              
        bool    bigger(const long inVal)                                   const { return( ((GetLong() + m_magnet) < inVal) ); }
        bool    smaller(const long inVal)                                  const { return( ((GetLong() - m_magnet) > inVal) ); }
        bool    equal(const long inVal)                                    const { return( (inVal <= (GetLong() + m_magnet)) && (inVal >= (GetLong() - m_magnet)) ); }
                                                                          
                                                                          
                                                                          
    // Additional operators interface                                     
    public:                                                               
        // dbl class object reference                                     
            bool        operator == (ea_dbl &inVal)                              { _sync(inVal); bool bEqual = equal(inVal.GetLong()); _unsync(inVal); return(bEqual); }
            bool        operator >  (ea_dbl &inVal)                              { _sync(inVal); bool bEqual = smaller(inVal.GetLong()); _unsync(inVal); return(bEqual); }
            bool        operator <  (ea_dbl &inVal)                              { _sync(inVal); bool bEqual = bigger(inVal.GetLong()); _unsync(inVal); return(bEqual); }
            bool        operator >= (ea_dbl &inVal)                              { _sync(inVal); bool bEqual = (smaller(inVal.GetLong()) || equal(inVal.GetLong())); _unsync(inVal); return(bEqual); }
            bool        operator <= (ea_dbl &inVal)                              { _sync(inVal); bool bEqual = (bigger(inVal.GetLong()) || equal(inVal.GetLong())); _unsync(inVal); return(bEqual); }
                                                                          
                                                                          
        // double as input                                                
            bool        operator == (const double inVal)                   const { return(equal(to_long(inVal))); }
            bool        operator >  (const double inVal)                   const { return(smaller(to_long(inVal))); }
            bool        operator <  (const double inVal)                   const { return(bigger(to_long(inVal))); }
            bool        operator >= (const double inVal)                   const { return(smaller(to_long(inVal)) || equal(to_long(inVal))); }
            bool        operator <= (const double inVal)                   const { return(bigger(to_long(inVal))  || equal(to_long(inVal))); }
                                                                          
        // Generic                                                        
            bool        operator !  ()                                     const { return(equal(0)); }
            void        operator ++ ()                                           { operator += (to_double(m_magnet)); }
            void        operator -- ()                                           { operator -= (to_double(m_magnet)); }
};



